# Some portions of this script are derived from these sources:
# http://www.danomagnum.com/wiki/Python3dCubeWorld
# http://blenderartists.org/forum/showthread.php?151185-Random-level-generator-%280-3-for-Blender-2-49%29

# Dungeon creation script
# Alt-p to execute
#
# author: Arnaud Couturier (piiichan)
# location: New Caledonia (Pacific island)
# date: March 2009 (Blender 2.48)
# version: 0.2
# license: free for any use
# rationale:
#       This script allows you to create dungeon-like levels

from panda3d.core import *
from direct.showbase.ShowBase import ShowBase
from direct.gui.OnscreenImage import OnscreenImage

import random
import math, sys, time, os
from FirstPersonCamera import FirstPersonCamera

loadPrcFileData("","sync-video #f")
loadPrcFileData('', 'win-size 1024 768')
loadPrcFileData("","win-origin -2 -2")

class Application(ShowBase):
    def __init__(self):
        ShowBase.__init__(self)
        self.accept("escape", sys.exit)
        base.win.setClearColor(Vec4(0.0,0.0,0.0,1)) # Background color
        base.setFrameRateMeter(True)
        render.setAntialias(AntialiasAttrib.MAuto)
   
        usecolor = True

        print "Starting.  Please wait..."
        StartTime = time.time()        
        
        d = DungeonGenerator(usecolor=usecolor)
        d.createMesh()
        d.createImage()
        self.player = d.player
        self.mapscale = (1/d.scale)*d.image_scale
        
        myDungeon = render.attachNewNode(d.geomnode)        

        EndTime = (time.time() - StartTime)/60
        minutes = int(EndTime)
        seconds = int((EndTime - float(minutes))*60)
        
        print "Geometry and octree done in %s minutes, %s seconds." %(minutes,seconds)
        print "Unflattened model has %s geometries." %(myDungeon.node().getNumGeoms())        
        print myDungeon.node()
        StartTime2 = time.time()
        
        print "Starting flatten."
        myDungeon.clearModelNodes()
        #myDungeon.flattenStrong()
        myDungeon.flattenMedium()
        #myDungeon.flattenLight()
        
        print "Flatten done."
        print "Flattened model has %s geometries." %(myDungeon.node().getNumGeoms())
        print myDungeon.node()
        
        if usecolor:
            for tex, bump in d.textureStages:
                myDungeon.setTexScale(tex, 2, 2, 1)
                myDungeon.setTexScale(bump, 2, 2, 1)            
            myMaterial = Material()
            myMaterial.setShininess(1.0) #Make this material shiny
            myDungeon.setMaterial(myMaterial)
        else:        
            tex = loader.loadTexture("models/dungeon/white_tile.png")
            myDungeon.setTexture(tex) 
            myDungeon.setTexScale(TextureStage.getDefault(), 3, 3, 1)
            tex = loader.loadTexture("models/dungeon/tile_bump3.jpg")
            ts = TextureStage("bump")
            ts.setMode(TextureStage.MNormal)
            myDungeon.setTexScale(ts, 3, 3, 1)
            myDungeon.setTexture(ts, tex)
            myMaterial = Material()
            myMaterial.setShininess(10.0) #Make this material shiny
            myDungeon.setMaterial(myMaterial)

        EndTime = (time.time() - StartTime2)/60
        minutes = int(EndTime)
        seconds = int((EndTime - float(minutes))*60)        
        print "Flatten done in %s minutes, %s seconds." %(minutes,seconds)
        EndTime = (time.time() - StartTime)/60
        minutes = int(EndTime)
        seconds = int((EndTime - float(minutes))*60)        
        print "All geometry done in %s minutes, %s seconds." %(minutes,seconds)
        
        self.handle_lights(0, 1)
        self.cam.setPos(1, 1, 3)        
        self.cam.node().getLens().setNearFar(0.01,150)
        self.cam.node().getLens().setFov(75)
        render.setShaderAuto()

        taskMgr.add(self.update_map, "mapTask")
        
        # Uses external FirstPersonCamera module
        self.mouseLook = FirstPersonCamera(self, self.cam, self.render, collisionHandler=None)
        self.mouseLook.start()
        self.accept("tab", self.mouseLook.stop)
        self.accept("`", self.mouseLook.start)
        self.mouseLook.movSens = 10
        self.mouseLook.setMat = False
        self.mouseLook.freeze_Z = True
        self.mouseLook.camZ = self.cam.getZ()
        self.mouseLook.accept('mouse1', setattr, [self.mouseLook, "forward", True])
        self.mouseLook.accept('mouse1-up', setattr, [self.mouseLook, "forward", False])
        self.mouseLook.accept('mouse3', setattr, [self.mouseLook, "backward", True])
        self.mouseLook.accept('mouse3-up', setattr, [self.mouseLook, "backward", False])
        # Rotate the camera
        self.mouseLook.camera.setH(self.mouseLook.refNode, self.mouseLook.camera.getH(self.mouseLook.refNode) - 45.0)

    def update_map(self, task):
        self.player.setX(self.cam.getX()*self.mapscale)
        self.player.setZ(-self.cam.getY()*self.mapscale)        
        return task.cont
        
    def handle_lights(self, IBL, point):
        if IBL:
            ambLight = AmbientLight("ambient")
            ambLight.setColor(Vec4(1.01, 1.01, 1.01, 1.0))
            ambNode = render.attachNewNode(ambLight)
            render.setLight(ambNode)

        if point:    
            pntLight = PointLight("point")
            pntLight.setColor(Vec4(2.0, 2.0, 2.0, 1.0))
            pntLight.setAttenuation((1, 0, 0.01))
            pntNode = render.attachNewNode(pntLight)
            pntNode.setPos(0, 2, 3)
            pntNode.reparentTo(self.cam)
            render.setLight(pntNode)
        

class DungeonGenerator:
    """This class generates dungeon-like levels."""
    # Most of these are Piichan's original functions, modified only to support multi-color handling.
    def __init__(self, width=300, length=300, floors=1, roomMinSize=5, roomMaxSize=20, totRooms=20, probaTurn=0.1, probaRoom=0.01, lightProba=0.01, usecolor=True, scale=5.0, height=2):

        #Room acceptable sizes
        self.roomWidthMin = roomMinSize
        self.roomWidthMax = roomMaxSize
        self.roomHeightMin = roomMinSize
        self.roomHeightMax = roomMaxSize

        #Constant values for cell types
        self.TYPE_WALL = 0x00000000
        self.TYPE_INSIDE = 0xffffffff
        self.TYPE_OUTSIDE = 0x66666666
        self.TYPE_SPECIAL = 0xffff0000

        # dungeon dimensions
        self.dungeonWidth = width
        self.dungeonLength = length
        self.dungeonHeight = floors

        # worm behavior
        self.roomsPerLvl = totRooms
        self.probaTurn = probaTurn
        self.probaRoom = probaRoom

        # rendering variables
        self.cellSize = 0.1
        self.wallHeight = 0.2
        
        # *************Variables added for Panda:***************************
        self.usecolor = usecolor
        self.scale = scale
        self.height = height
        self.image_scale = 2.0

        self.colors = [None for i in range(8)]
        #self.colors[0] = (0.0,0.0,0.0, 1)#(0,0,0) #black
        self.colors[0] = (0.5,0.5,0.5, 1)#(128,128,128) #gray for black
        self.colors[1] = (1.0,1.0,1.0, 1)#(255,255,255) #white
        self.colors[2] = (1.0,0.0,0.0, 1)#(255,0,0) #red
        self.colors[3] = (0.0,1.0,0.0, 1)#(0,255,0) #green
        self.colors[4] = (0.0,0.0,1.0, 1)#(0,0,255) #blue
        self.colors[5] = (1.0,1.0,0.0, 1)#(255,255,0) #yellow
        self.colors[6] = (0.75,0.0,0.75, 1)#(192,0,192) #purple
        self.colors[7] = (1.0,0.5,0.0, 1)#(255,128,0) #orange        
        self.colornames = ['black','white','red','green','blue','yellow','purple','orange']

        self.makeTextures()
        # **************** End Panda variables ******************************

        # TODO: place following code in a function "createDungeon"
        #The data structure that holds the dungeon level data
        self.dat = [[[self.TYPE_OUTSIDE for k in range(self.dungeonHeight)] for j in range(self.dungeonLength)] for i in range(self.dungeonWidth)]
        self.pixels = [[[-1 for k in range(self.dungeonHeight)] for j in range(self.dungeonLength)] for i in range(self.dungeonWidth)]
        
        # generate corridors & place rooms
        rooms = self.worm()
        for i in range(len(rooms)):
            self.carveRoom(rooms[i][0], rooms[i][1], rooms[i][2])

########################################################################################
    def colorNode(self):
        """Set the texture and normal maps"""
        color_attribs = []
        self.textureStages = []
        for i in range(len(self.colors)):            
            name = self.colornames[i]
            tex = loader.loadTexture("models/dungeon/%s_tile.png" %(name))
            texAttrib = TextureAttrib.make()            
            ts_tex = TextureStage("%s_tex"%(name))
            ts_tex.setMode(TextureStage.MModulate)                
            texAttrib = texAttrib.addOnStage(ts_tex, tex)
            bump = loader.loadTexture("models/dungeon/tile_bump3.jpg")
            ts_bump = TextureStage("%s_bump"%(name))
            ts_bump.setMode(TextureStage.MNormal)
            texAttrib = texAttrib.addOnStage(ts_bump, bump)
            color_attribs.append(texAttrib)
            self.textureStages.append([ts_tex,ts_bump])
            
        for index in range(self.geomnode.getNumGeoms()):
            color = self.color_lookup[index]
            texAttrib = color_attribs[color]
            newRenderState = self.geomnode.getGeomState(index).addAttrib(texAttrib,1)
            self.geomnode.setGeomState(index, newRenderState)

########################################################################################
    def createImage(self):
        """Set up the dungeon map for display"""
        img = PNMImage(self.dungeonWidth, self.dungeonLength)
        img.addAlpha()
        img.alphaFill()
        for x in range(self.dungeonWidth):
            for y in range(self.dungeonLength):
                if self.dat[x][y][0] == self.TYPE_INSIDE:                    
                    colorVec = self.colors[self.pixels[x][y][0]]
                    img.setXelA(x,y,colorVec)
        #img.write("test.png")
        imgText = Texture()
        imgText.load(img)        
        imageObject = OnscreenImage()
        imageObject.setImage(imgText)        
        imageObject.reparentTo(pixel2d)          
        w = (self.dungeonWidth/2)*self.image_scale
        l = (self.dungeonLength/2)*self.image_scale
        imageObject.setPos(w,0,-l)
        imageObject.setScale(w,1,l)                
        imageObject.setTransparency(TransparencyAttrib.MAlpha)
        self.drawPlayer()
        self.map = imageObject

    def drawPlayer(self,size=300):
        """Create the player sprite"""
        img = PNMImage(size, size)
        for i in range(size):
            for j in range(size): # Set up a multi-color pattern, to make player sprite appear to flash               
                if i==0 or j==0 or i==size-1 or j==size-1:
                    img.setXel(i,j,(1,1,1))
                else:
                    if i%2 == 0:
                        img.setXel(i,j,(0,1,0))
                    else:
                        img.setXel(i,j,(1,0,0))
        imgText = Texture()
        imgText.load(img)
        imageObject = OnscreenImage()
        imageObject.setImage(imgText)        
        imageObject.reparentTo(pixel2d)
        self.player = imageObject

    def makeTextures(self, size=240):
        colors = [(0, 0, 0),
                  (255, 255, 255),
                  (185, 0, 0),
                  (81, 190, 0),
                  (0, 99, 192),
                  (187, 181, 0),
                  (96, 0, 186),
                  (187, 97, 0)]
        path = "models/dungeon/"
        path.replace("/", os.sep)
        for i in range(len(colors)):
            color = colors[i]
            name = self.colornames[i]
            texture_path = os.path.join(path, "%s_tile.png" %(name))
            if os.path.exists(texture_path):
                print "Texture %s already exists." %(name)
                continue
            print "Generating texture %s..." %(name)
            if name == "white":
                color2 = (0, 0, 0)
            else:
                color2 = (255, 255, 255)
            img = PNMImage(size, size)
            for w in range(size):
                for h in range(size):
                    if w % 30 == 0 or h % 30 == 0 or w == size-1 or h == size-1:
                        img.setXelVal(w,h,color2)
                    else:
                        img.setXelVal(w,h,color)
            img.gaussianFilter()
            img.boxFilter()
            img.write(texture_path)        
        
########################################################################################
    def worm(self):
        """Carves the dungeon like a worm.
        Returns a list of positions (x,y) inside the dungeon where rooms are to be created later."""
        rooms = [[0,0,0]]
        self.dat[0][0][0] = self.TYPE_SPECIAL
        wormX = 0
        wormY = 0
        usecolor = self.usecolor

        if usecolor is True:
            color = random.randint(0,7) #Random color from the palette
        else:
            color = 1

        # for every level in the building (z-axis)
        for wormZ in range(self.dungeonHeight):
            lvlRooms = 1
            wormDirection = [1,0]            
            self.randomizeDirection(wormDirection)

            # carve first room at starting point
            self.dat[wormX][wormY][wormZ] = self.TYPE_SPECIAL
            rooms.append([wormX, wormY, wormZ])

            # carve current dungeon level
            while lvlRooms < self.roomsPerLvl:
                nextX = wormX + wormDirection[0]
                nextY = wormY + wormDirection[1]
                # if next move is not valid or if the worm decides to change its direction: change direction
                if nextX < 0 or self.dungeonWidth <= nextX or nextY < 0 or self.dungeonLength <= nextY or random.random() < self.probaTurn:
                    self.randomizeDirection(wormDirection)
                    nextX = wormX + wormDirection[0]
                    nextY = wormY + wormDirection[1]
                    if usecolor is True and random.random() < 0.2:
                        color = random.randint(0,7)
                # if after changing the direction, the next move is still not valid, the worm must face the opposite direction
                if nextX < 0 or self.dungeonWidth <= nextX or nextY < 0 or self.dungeonLength <= nextY:
                    wormDirection[0] = -wormDirection[0]
                    wormDirection[1] = -wormDirection[1]
                    nextX = wormX + wormDirection[0]
                    nextY = wormY + wormDirection[1]
                    if usecolor is True and random.random() < 0.2:
                        color = random.randint(0,7)

                # move + carve
                wormX = nextX
                wormY = nextY                
                self.dat[wormX][wormY][wormZ] = self.TYPE_INSIDE
                if usecolor is True:
                    self.pixels[wormX][wormY][wormZ] = color

                # place room
                if random.random() < self.probaRoom:
                    rooms.append([wormX, wormY, wormZ])
                    lvlRooms += 1
                    self.dat[wormX][wormY][wormZ] = self.TYPE_SPECIAL
                    self.pixels[wormX][wormY][wormZ] = color
        return rooms

########################################################################################
    def randomizeDirection(self, initDirection):
        """Returns a random direction. If the original direction is vertical, the result is horizontal and vice versa."""
        if initDirection[0] != 0: #was along x
            initDirection[0] = 0
            if random.random() >= 0.5:
                initDirection[1] = -1
            else:
                initDirection[1] = 1
        else: # was along y
            initDirection[1] = 0
            if random.random() >= 0.5:
                initDirection[0] = -1
            else:
                initDirection[0] = 1

########################################################################################
    # Create a room of random size in the acceptable room size range.
    def carveRoom(self, centerX, centerY, centerZ):
        usecolor = self.usecolor
        x1 = int(Utils.getRandomInt(max(0, centerX-self.roomWidthMin/2), max(0, centerX-self.roomWidthMax/2)))
        y1 = int(Utils.getRandomInt(max(0, centerY-self.roomHeightMin/2), max(0, centerY-self.roomHeightMax/2)))

        x2 = int(Utils.getRandomInt(min(self.dungeonWidth-1, centerX + self.roomWidthMin/2), min(self.dungeonWidth-1, centerX + self.roomWidthMax/2)))
        y2 = int(Utils.getRandomInt(min(self.dungeonLength-1, centerY + self.roomHeightMin/2), min(self.dungeonLength-1, centerY + self.roomHeightMax/2)))
        color = 0
        if usecolor is True:
            color = self.pixels[centerX][centerY][centerZ]
            if random.random() < 0.2:
                color = random.randint(0,7)

        # inside
        for i in range(x1, x2):
            for j in range(y1, y2):
                self.dat[i][j][centerZ] = self.TYPE_INSIDE
                if usecolor == True:
                    self.pixels[i][j][centerZ] = color

########################################################################################
    def createMesh(self, createRoof=True):
        """
        Adapted from Piichan's Blender script and Panda3dcubes.  Heavily modified.
        """
        if self.usecolor:
            self.color_lookup = []
        usecolor = self.usecolor
        scale = self.scale
        height = self.height
        self.geomnode = GeomNode('dungeon')
        for z in range(self.dungeonHeight):
            for x in range(self.dungeonWidth):
                for y in range(self.dungeonLength):
                    lvlHeight = z * self.wallHeight
                    # create floors and  walls (walls only at the edge of inside and outside)
                    if self.dat[x][y][z] == self.TYPE_INSIDE:
                        
                        # floor face
                        bottom = makeSquare(x*scale ,(y+1)*scale ,z*scale*height ,(x+1)*scale ,y*scale ,z*scale*height, reverse=1) 
                        self.geomnode.addGeom(bottom)
                        if usecolor:
                            self.color_lookup.append(self.pixels[x][y][z])
                        
                        # ceiling face
                        if createRoof:
                            top    = makeSquare(x*scale ,(y+1)*scale ,(z+1)*scale*height ,(x+1)*scale ,y*scale ,(z+1)*scale*height)
                                                       
                            self.geomnode.addGeom(top)
                            if usecolor:
                                self.color_lookup.append(self.pixels[x][y][z])
                            
                        # west(left) wall intact: build it
                        if (0 < x and self.dat[x-1][y][z] == self.TYPE_OUTSIDE) or x <= 0:
                            right  = makeSquare(x*scale ,y*scale ,z*scale ,x*scale ,(y+1)*scale ,(z+1)*scale, reverse=0, height=height, scale=scale, floor=z)                            
                            self.geomnode.addGeom(right)
                            if usecolor:
                                self.color_lookup.append(self.pixels[x][y][z])

                        # east(right) wall intact: build it
                        if (x < self.dungeonWidth-1 and self.dat[x+1][y][z] == self.TYPE_OUTSIDE) or x >= self.dungeonWidth-1:                            
                            left   = makeSquare((x+1)*scale ,y*scale ,z*scale ,(x+1)*scale ,(y+1)*scale ,(z+1)*scale, reverse=1, height=height, scale=scale, floor=z)                            
                            self.geomnode.addGeom(left)
                            if usecolor:
                                self.color_lookup.append(self.pixels[x][y][z])

                        # south wall intact: build it
                        if (0 < y and self.dat[x][y-1][z] == self.TYPE_OUTSIDE) or y <= 0:
                            front  = makeSquare(x*scale ,y*scale ,z*scale ,(x+1)*scale ,y*scale ,(z+1)*scale, reverse=1, height=height, scale=scale, floor=z)                            
                            self.geomnode.addGeom(front)
                            if usecolor:
                                self.color_lookup.append(self.pixels[x][y][z])

                        # north wall intact: build it
                        if (y < self.dungeonLength-1 and self.dat[x][y+1][z] == self.TYPE_OUTSIDE) or y >= self.dungeonLength-1:                            
                            back   = makeSquare(x*scale ,(y+1)*scale ,z*scale ,(x+1)*scale ,(y+1)*scale ,(z+1)*scale, height=height, scale=scale, floor=z)                            
                            self.geomnode.addGeom(back)
                            if usecolor:
                                self.color_lookup.append(self.pixels[x][y][z])
            if self.usecolor:
                self.colorNode()

########################################################################################
#you cant normalize in-place so this is a helper function
def myNormalize(myVec):
    myVec.normalize()
    return myVec

def getNorms(v1, v2, v3):
    e1 = myNormalize(v1 - v2)
    e2 = myNormalize(v3 - v2)
    norm = myNormalize(e2.cross(e1))
    return norm

def getBinorms(edge, norm):
    tan = myNormalize(norm.cross(edge))
    binorm = myNormalize(tan.cross(norm))
    return tan, binorm

#helper function to make a square given the Lower-Left-Hand and Upper-Right-Hand corners
def makeSquare(x1,y1,z1, x2,y2,z2, reverse=0, height=1, scale=1.0, floor=0):
    """
    This started as the Panda Procedural Cube example.  Heavily modified.
    """
    # Build array for new format.
    array = GeomVertexArrayFormat()
    array.addColumn(InternalName.make('vertex'), 3, Geom.NTFloat32, Geom.CPoint)    
    array.addColumn(InternalName.make('texcoord'), 2, Geom.NTFloat32, Geom.CTexcoord)
    array.addColumn(InternalName.make('normal'), 3, Geom.NTFloat32, Geom.CVector)
    array.addColumn(InternalName.make('binormal'), 3, Geom.NTFloat32, Geom.CVector)
    array.addColumn(InternalName.make('tangent'), 3, Geom.NTFloat32, Geom.CVector)

    # Create and register format.
    format = GeomVertexFormat()
    format.addArray(array)
    format = GeomVertexFormat.registerFormat(format)

    # Create vdata and vertex writers.
    vdata=GeomVertexData('square', format, Geom.UHStatic)

    vertex = GeomVertexWriter(vdata, 'vertex')
    normal = GeomVertexWriter(vdata, 'normal')
    texcoord = GeomVertexWriter(vdata, 'texcoord')
    tangent = GeomVertexWriter(vdata, 'tangent')
    binormal = GeomVertexWriter(vdata, 'binormal')

    old_z1 = z1
    old_z2 = z2
    for lvl in range(height):
        z1 = old_z1 + (lvl + floor) *scale
        z2 = old_z2 + (lvl + floor) *scale
        #make sure we draw the square in the right plane
        if x1!=x2:
            
            if reverse:
                vertex.addData3f(x1, y2, z2) #4
                vertex.addData3f(x2, y2, z2) #3
                vertex.addData3f(x2, y1, z1) #2
                vertex.addData3f(x1, y1, z1) #1
                
                v1 = Vec3(x1, y2, z2) # Get the vertices
                v2 = Vec3(x2, y2, z2)
                v3 = Vec3(x2, y1, z1)
                v4 = Vec3(x1, y1, z1)

                norm = getNorms(v1, v2, v3)
                edge = myNormalize(v2 - v1)
                tan, binorm = getBinorms(edge, norm)
                for i in range(4):
                    normal.addData3f(norm)
                    binormal.addData3f(binorm)
                    tangent.addData3f(-tan)
                
            else:        
                vertex.addData3f(x1, y1, z1)
                vertex.addData3f(x2, y1, z1)
                vertex.addData3f(x2, y2, z2)
                vertex.addData3f(x1, y2, z2)
                
                v1 = Vec3(x1, y1, z1)
                v2 = Vec3(x2, y1, z1)
                v3 = Vec3(x2, y2, z2)
                v4 = Vec3(x1, y2, z2)
                
                norm = getNorms(v1, v2, v3) # ceiling
                edge = myNormalize(v2 - v1)
                tan, binorm = getBinorms(edge, norm)
                for i in range(4):
                    normal.addData3f(norm)
                    binormal.addData3f(binorm)
                    tangent.addData3f(tan)
                    
        else:        
            if reverse:
                vertex.addData3f(x1, y1, z2) #4
                vertex.addData3f(x2, y2, z2) #3
                vertex.addData3f(x2, y2, z1) #2
                vertex.addData3f(x1, y1, z1) #1
                
                v1 = Vec3(x1, y1, z2)
                v2 = Vec3(x2, y2, z2)
                v3 = Vec3(x2, y2, z1)
                v4 = Vec3(x1, y1, z1)

                norm = getNorms(v1, v2, v3)
                edge = myNormalize(v2 - v1) 
                tan, binorm = getBinorms(edge, norm)
                for i in range(4):
                    normal.addData3f(norm)
                    binormal.addData3f(binorm)
                    tangent.addData3f(-tan)
            else:
            
                vertex.addData3f(x1, y1, z1)
                vertex.addData3f(x2, y2, z1)
                vertex.addData3f(x2, y2, z2)
                vertex.addData3f(x1, y1, z2)
                
                v1 = Vec3(x1, y1, z1)
                v2 = Vec3(x2, y2, z1)
                v3 = Vec3(x2, y2, z2)
                v4 = Vec3(x1, y1, z2)

                norm = getNorms(v1, v2, v3)
                edge = myNormalize(v2 - v1)
                tan, binorm = getBinorms(edge, norm)
                for i in range(4):
                    normal.addData3f(norm)
                    binormal.addData3f(binorm)
                    tangent.addData3f(tan)
              
        if reverse:
            texcoord.addData2f(1.0, 1.0) #4
            texcoord.addData2f(1.0, 0.0) #3
            texcoord.addData2f(0.0, 0.0) #2
            texcoord.addData2f(0.0, 1.0) #1        
            
        else:    
            texcoord.addData2f(0.0, 1.0)
            texcoord.addData2f(0.0, 0.0)
            texcoord.addData2f(1.0, 0.0)
            texcoord.addData2f(1.0, 1.0)

        #quads arent directly supported by the Geom interface
        #you might be interested in the CardMaker class if you are
        #interested in rectangle though
        tri1=GeomTriangles(Geom.UHDynamic)
        tri2=GeomTriangles(Geom.UHDynamic)

        tri1.addVertex(0+(lvl*4))
        tri1.addVertex(1+(lvl*4))
        tri1.addVertex(3+(lvl*4))

        tri2.addVertex(1+(lvl*4))
        tri2.addVertex(2+(lvl*4))
        tri2.addVertex(3+(lvl*4))

        tri1.closePrimitive()
        tri2.closePrimitive()
        if lvl == 0:
            tris = []
        tris.append(tri1)
        tris.append(tri2)

    square=Geom(vdata)
    for tri in tris:
        square.addPrimitive(tri)
    return square
    
########################################################################################
class Utils:
    @staticmethod
    def getRandomInt(minInclusive, maxInclusive):
        return math.floor(random.random() * (maxInclusive - minInclusive + 1)) + minInclusive


if __name__ == "__main__":
  gameApp = Application()
  gameApp.run()
